@@ -1,4 +1,5 @@
 /*
+ * Copyright © 2024 Thomas E. Dickey
  * Copyright © 2002 Keith Packard
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
@@ -30,7 +31,7 @@ XcursorCursorsCreate (Display *dpy, int size)
     XcursorCursors  *cursors;
 
     cursors = malloc (sizeof (XcursorCursors) +
-		      size * sizeof (Cursor));
+		      (size_t) size * sizeof (Cursor));
     if (!cursors)
 	return NULL;
     cursors->ref = 1;
@@ -91,6 +92,7 @@ XcursorAnimateNext (XcursorAnimate *animate)
     return cursor;
 }
 
+#if RENDER_MAJOR > 0 || RENDER_MINOR >= 5
 static int
 nativeByteOrder (void)
 {
@@ -98,6 +100,7 @@ nativeByteOrder (void)
 
     return (*((char *) &x) == 1) ? LSBFirst : MSBFirst;
 }
+#endif
 
 static XcursorUInt
 _XcursorPixelBrightness (XcursorPixel p)
@@ -124,7 +127,7 @@ _XcursorDivideAlpha (XcursorUInt value, XcursorUInt alpha)
     value = value * 255 / alpha;
     if (value > 255)
 	value = 255;
-    return value | (value << 8);
+    return (unsigned short) (value | (value << 8));
 }
 
 static void
@@ -144,7 +147,7 @@ _XcursorPixelToColor (XcursorPixel p, XColor *color)
 static void
 _XcursorDumpImage (XImage *image)
 {
-    FILE    *f = fopen ("/tmp/images", "a");
+    FILE    *f = fopen ("/tmp/images", "a" FOPEN_CLOEXEC);
     int	    x, y;
     if (!f)
 	return;
@@ -162,7 +165,7 @@ _XcursorDumpImage (XImage *image)
 static void
 _XcursorDumpColor (XColor *color, char *name)
 {
-    FILE    *f = fopen ("/tmp/images", "a");
+    FILE    *f = fopen ("/tmp/images", "a" FOPEN_CLOEXEC);
     fprintf (f, "%s: %x %x %x\n", name,
 	     color->red, color->green, color->blue);
     fflush (f);
@@ -194,6 +197,8 @@ _XcursorCompareBlue (const void *a, const void *b)
     return (int) (((*ap >> 0) & 0xff) - ((*bp >> 0) & 0xff));
 }
 
+#define ScaledPixels(c,n) ((c)/(XcursorPixel)(n))
+
 static XcursorPixel
 _XcursorAverageColor (XcursorPixel *pixels, int npixels)
 {
@@ -212,7 +217,10 @@ _XcursorAverageColor (XcursorPixel *pixels, int npixels)
 	green += (p >> 8) & 0xff;
 	blue += (p >> 0) & 0xff;
     }
-    return (0xffU << 24) | ((red/npixels) << 16) | ((green/npixels) << 8) | (blue/npixels);
+    return (0xffU << 24)
+	    | (ScaledPixels(red, npixels) << 16)
+	    | (ScaledPixels(green, npixels) << 8)
+	    |  ScaledPixels(blue, npixels);
 }
 
 typedef struct XcursorCoreCursor {
@@ -226,7 +234,7 @@ static Bool
 _XcursorHeckbertMedianCut (const XcursorImage *image, XcursorCoreCursor *core)
 {
     XImage	    *src_image = core->src_image, *msk_image = core->msk_image;
-    int		    npixels = image->width * image->height;
+    unsigned int    npixels = image->width * image->height;
     int		    ncolors;
     int		    n;
     XcursorPixel    *po, *pn, *pc;
@@ -258,7 +266,7 @@ _XcursorHeckbertMedianCut (const XcursorImage *image, XcursorCoreCursor *core)
     pc = colors;
     max_blue = max_green = max_red = 0;
     min_blue = min_green = min_red = 255;
-    n = npixels;
+    n = (int) npixels;
     while (n--)
     {
 	p = *po++;
@@ -285,7 +293,7 @@ _XcursorHeckbertMedianCut (const XcursorImage *image, XcursorCoreCursor *core)
 	    p = 0;
 	*pn++ = p;
     }
-    ncolors = pc - colors;
+    ncolors = (int) (pc - colors);
 
     /*
      * Compute longest dimension and sort
@@ -297,7 +305,7 @@ _XcursorHeckbertMedianCut (const XcursorImage *image, XcursorCoreCursor *core)
 	compare = _XcursorCompareRed;
     else
 	compare = _XcursorCompareBlue;
-    qsort (colors, ncolors, sizeof (XcursorPixel), compare);
+    qsort (colors, (size_t) ncolors, sizeof (XcursorPixel), compare);
     /*
      * Compute average colors on both sides of the cut
      */
@@ -309,8 +317,8 @@ _XcursorHeckbertMedianCut (const XcursorImage *image, XcursorCoreCursor *core)
      * Select best color for each pixel
      */
     pn = pixels;
-    for (y = 0; y < image->height; y++)
-	for (x = 0; x < image->width; x++)
+    for (y = 0; (unsigned) y < image->height; y++)
+	for (x = 0; (unsigned) x < image->width; x++)
 	{
 	    p = *pn++;
 	    if (p & 0xff000000)
@@ -359,12 +367,12 @@ _XcursorBayerOrderedDither (const XcursorImage *image, XcursorCoreCursor *core)
     XcursorPixel    a, i, d;
 
     pixel = image->pixels;
-    for (y = 0; y < image->height; y++)
-	for (x = 0; x < image->width; x++)
+    for (y = 0; (unsigned) y < image->height; y++)
+	for (x = 0; (unsigned) x < image->width; x++)
 	{
 	    p = *pixel++;
-	    a = ((p >> 24) * DITHER_SIZE + 127) / 255;
-	    i = (_XcursorPixelBrightness (p) * DITHER_SIZE + 127) / 255;
+	    a = (XcursorPixel) (((p >> 24) * DITHER_SIZE + 127) / 255);
+	    i = (XcursorPixel) ((_XcursorPixelBrightness (p) * DITHER_SIZE + 127) / 255);
 	    d = orderedDither[y&(DITHER_DIM-1)][x&(DITHER_DIM-1)];
 	    if (a > d)
 	    {
@@ -395,12 +403,12 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
     int		    *aPicture, *iPicture, *aP, *iP;
     XcursorPixel    *pixel, p;
     int		    aR, iR, aA, iA;
-    int		    npixels = image->width * image->height;
+    unsigned int    npixels = image->width * image->height;
     int		    n;
     int		    right = 1;
-    int		    belowLeft = image->width - 1;
-    int		    below = image->width;
-    int		    belowRight = image->width + 1;
+    int		    belowLeft = (int) (image->width - 1);
+    int		    below = (int) image->width;
+    int		    belowRight = (int) (image->width + 1);
     int		    iError, aError;
     int		    iErrorRight, aErrorRight;
     int		    iErrorBelowLeft, aErrorBelowLeft;
@@ -420,10 +428,10 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
     pixel = image->pixels;
     iP = iPicture;
     aP = aPicture;
-    n = npixels;
+    n = (int) npixels;
     max_inten = 0;
     min_inten = 0xff;
-    while (n--)
+    while (n-- > 0)
     {
 	p = *pixel++;
 	*aP++ = (int) (p >> 24);
@@ -438,8 +446,8 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
     iP = iPicture;
     aP = aPicture;
     mean_inten = (max_inten + min_inten + 1) >> 1;
-    for (y = 0; y < image->height; y++)
-	for (x = 0; x < image->width; x++)
+    for (y = 0; (unsigned) y < image->height; y++)
+	for (x = 0; (unsigned) x < image->width; x++)
 	{
 	    aR = *aP;
 	    iR = *iP;
@@ -475,12 +483,12 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
 	    aErrorBelow = (aError * 5) >> 4;
 	    aErrorBelowRight = (aError - aErrorRight -
 				aErrorBelowLeft - aErrorBelow);
-	    if (x < image->width - 1)
+	    if (x < ((int)image->width - 1))
 	    {
 		iP[right] += iErrorRight;
 		aP[right] += aErrorRight;
 	    }
-	    if (y < image->height - 1)
+	    if (y < ((int)image->height - 1))
 	    {
 		if (x)
 		{
@@ -489,7 +497,7 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
 		}
 		iP[below] += iErrorBelow;
 		aP[below] += aErrorBelow;
-		if (x < image->width - 1)
+		if (x < ((int)image->width - 1))
 		{
 		    iP[belowRight] += iErrorBelowRight;
 		    aP[belowRight] += aErrorBelowRight;
@@ -501,10 +509,10 @@ _XcursorFloydSteinberg (const XcursorImage *image, XcursorCoreCursor *core)
     free (iPicture);
     core->on_color.red =
     core->on_color.green =
-    core->on_color.blue = (min_inten | min_inten << 8);
+    core->on_color.blue = (unsigned short) (min_inten | min_inten << 8);
     core->off_color.red =
     core->off_color.green =
-    core->off_color.blue = (max_inten | max_inten << 8);
+    core->off_color.blue = (unsigned short) (max_inten | max_inten << 8);
     return True;
 }
 
@@ -518,8 +526,8 @@ _XcursorThreshold (const XcursorImage *image, XcursorCoreCursor *core)
      * Draw the image, picking black for dark pixels and white for light
      */
     pixel = image->pixels;
-    for (y = 0; y < image->height; y++)
-	for (x = 0; x < image->width; x++)
+    for (y = 0; (unsigned) y < image->height; y++)
+	for (x = 0; (unsigned) x < image->width; x++)
 	{
 	    p = *pixel++;
 	    if ((p >> 24) >= 0x80)
@@ -560,8 +568,8 @@ XcursorImageLoadCursor (Display *dpy, const XcursorImage *image)
 	GC		    gc;
 	XRenderPictFormat   *format;
 
-	ximage.width = image->width;
-	ximage.height = image->height;
+	ximage.width = (int) image->width;
+	ximage.height = (int) image->height;
 	ximage.xoffset = 0;
 	ximage.format = ZPixmap;
 	ximage.data = (char *) image->pixels;
@@ -571,7 +579,7 @@ XcursorImageLoadCursor (Display *dpy, const XcursorImage *image)
 	ximage.bitmap_pad = 32;
 	ximage.depth = 32;
 	ximage.bits_per_pixel = 32;
-	ximage.bytes_per_line = image->width * 4;
+	ximage.bytes_per_line = (int) (image->width * 4);
 	ximage.red_mask = 0xff0000;
 	ximage.green_mask = 0x00ff00;
 	ximage.blue_mask = 0x0000ff;
@@ -608,12 +616,12 @@ XcursorImageLoadCursor (Display *dpy, const XcursorImage *image)
 				       0, NULL, image->width, image->height,
 				       32, 0);
 	core.src_image->data = Xmalloc (image->height *
-					core.src_image->bytes_per_line);
+					(unsigned) core.src_image->bytes_per_line);
 	core.msk_image = XCreateImage (dpy, NULL, 1, ZPixmap,
 				       0, NULL, image->width, image->height,
 				       32, 0);
 	core.msk_image->data = Xmalloc (image->height *
-					core.msk_image->bytes_per_line);
+					(unsigned) core.msk_image->bytes_per_line);
 
 	switch (info->dither) {
 	case XcursorDitherThreshold:
@@ -708,7 +716,7 @@ XcursorImagesLoadCursor (Display *dpy, const XcursorImages *images)
 
 	if (!cursors)
 	    return 0;
-	anim = malloc (cursors->ncursor * sizeof (XAnimCursor));
+	anim = malloc ((size_t) cursors->ncursor * sizeof (XAnimCursor));
 	if (!anim)
 	{
 	    XcursorCursorsDestroy (cursors);
@@ -735,9 +743,11 @@ Cursor
 XcursorFilenameLoadCursor (Display *dpy, const char *file)
 {
     int		    size = XcursorGetDefaultSize (dpy);
-    XcursorImages   *images = XcursorFilenameLoadImages (file, size);
+    XcursorBool     resize = XcursorGetResizable (dpy);
+    XcursorImages   *images;
     Cursor	    cursor;
 
+    images = _XcursorFilenameLoadImages (file, size, resize);
     if (!images)
 	return None;
     cursor = XcursorImagesLoadCursor (dpy, images);
@@ -749,9 +759,11 @@ XcursorCursors *
 XcursorFilenameLoadCursors (Display *dpy, const char *file)
 {
     int		    size = XcursorGetDefaultSize (dpy);
-    XcursorImages   *images = XcursorFilenameLoadImages (file, size);
+    XcursorBool     resize = XcursorGetResizable (dpy);
+    XcursorImages   *images;
     XcursorCursors  *cursors;
 
+    images = _XcursorFilenameLoadImages (file, size, resize);
     if (!images)
 	return NULL;
     cursors = XcursorImagesLoadCursors (dpy, images);
@@ -777,11 +789,11 @@ _XcursorCreateGlyphCursor(Display	    *dpy,
 
     LockDisplay(dpy);
     GetReq(CreateGlyphCursor, req);
-    cid = req->cid = XAllocID(dpy);
-    req->source = source_font;
-    req->mask = mask_font;
-    req->sourceChar = source_char;
-    req->maskChar = mask_char;
+    cid = req->cid = (CARD32) XAllocID(dpy);
+    req->source = (CARD32) source_font;
+    req->mask = (CARD32) mask_font;
+    req->sourceChar = (CARD16) source_char;
+    req->maskChar = (CARD16) mask_char;
     req->foreRed = foreground->red;
     req->foreGreen = foreground->green;
     req->foreBlue = foreground->blue;
@@ -800,8 +812,10 @@ _XcursorCreateGlyphCursor(Display	    *dpy,
 Cursor
 _XcursorCreateFontCursor (Display *dpy, unsigned int shape)
 {
-    static XColor _Xconst foreground = { 0,    0,     0,     0  };  /* black */
-    static XColor _Xconst background = { 0, 65535, 65535, 65535 };  /* white */
+#define DATA(c) { 0UL, c, c, c, 0, 0 }
+    static XColor _Xconst foreground = DATA(0);  /* black */
+    static XColor _Xconst background = DATA(65535);  /* white */
+#undef DATA
 
     /*
      * the cursor font contains the shape glyph followed by the mask
